#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <string.h>
#include <unistd.h>
#include <SDL.h>
#include <SDLttf.h>
#include <SDL_audio.h>
#include <SDL_error.h>
#include "mixer.h"

#define MAXWORDSIZE 100
#define MAXSTATE 7
#define NUMSOUNDS 3

enum {WINNER, LOOSER};
int soundflag;
SDL_Surface *screen, *blank,*blackblank, *title_image;
SDL_Surface *sound_on, *sound_off, *hangman[MAXSTATE];
SDL_Rect dest;
TTF_Font *font;
SDL_Color white = { 0xFF, 0xFF, 0xFF, 0 };
SDL_Color black = { 0x00, 0x00, 0x00, 0 };
Mix_Chunk *good_sounds[NUMSOUNDS];
Mix_Chunk *bad_sounds[NUMSOUNDS];

struct word {
   char letter;
   int flag;
   struct word *next;
   SDL_Surface *image;
};

void myfree(struct word *theword) {
   if (theword->next != NULL) myfree(theword->next);
   SDL_FreeSurface(theword->image);
   free(theword);
   theword=NULL;
}

int nonchar(char a) {
   if (a >= 'a' && a <='z') return 0;
   if (a >= 'A' && a <='Z') return 0;
   return 1;
}

struct word *get_word(char *dictionary, struct word *top) {
   FILE *myfile;
   struct word *current;
   char tmpword[MAXWORDSIZE],tmp[4];
   int x,size,which;

   if ((myfile = fopen(dictionary,"r")) == NULL) { return 0; }
   fscanf(myfile,"%d",&size);
   which=1+ (int)(rand() % size);
   for (x=0;x<=which;x++) {
      fgets(tmpword,MAXWORDSIZE,myfile);
   }
   current = top;
   x=0;
   while (tmpword[x] != '\n') {
      if (x==0) {
          top = (struct word *)malloc(sizeof(struct word));
          current = top;
      } else {
         current->next = (struct word *)malloc(sizeof(struct word));
         current = current->next;
      }
      current->letter = tmpword[x];
      if (nonchar(tmpword[x])) {
         current->flag=1;
      } else {
         current->flag=0;
      }
      sprintf(tmp,"%c",current->letter);
      current->image =  TTF_RenderText(font,tmp,white,black);
      if (current->image == NULL) {
         fprintf(stderr,"Problem rendering font. %s\n",SDL_GetError());
         exit(2);
      }
      current->next=NULL;
      x++;
   }
   current = top;
   close((int)myfile);
   return top;
}

void draw_text(SDL_Surface *text,int a,int b) {

   dest.x = a;
   dest.y = b - text->h;
   dest.w = text->w;
   dest.h = text->h;
   if (SDL_BlitSurface(text, NULL, screen, &dest) < 0) {
      fprintf(stderr,"Couldn't blit text to display: %s\n",SDL_GetError());
      exit(2);
   }
   SDL_UpdateRects(screen,1,&dest);
}

void draw_image(SDL_Surface *image,int x,int y) {
   SDL_SetClipping(image, 0, 0, screen->h, screen->w);

   dest.x = x;
   dest.y = y;
   dest.w = image->w;
   dest.h = image->h;
   if ( SDL_BlitSurface(image, NULL, screen, &dest) < 0 ) {
      SDL_FreeSurface(image);
      fprintf(stderr,"Unable to blit\n");
      exit(1);
   }
   SDL_UpdateRects(screen,1,&dest);
}

void draw_display(int cstate,struct word *aword) {
   struct word *cletter;
   char temp[2];
   int x=200;
   int y=300;

   cletter=aword;
   /* draw state picture */
   if ( cstate < MAXSTATE ) draw_image((SDL_Surface *)hangman[cstate],5,50);
   while (cletter!=NULL) {
      x=x+35;
      if (x>560) {
         x=200;
         y+=50;
      }
      if (cletter->flag==1) {
 	 draw_text(blackblank,x,y);
         switch(cletter->letter) {
            case 'p':
            case 'g':
            case 'y':
  	       draw_text(cletter->image,x,y+8);
               break;
            default:
  	       draw_text(cletter->image,x,y);
               break;
         }
      } else {
	 draw_text(blank,x,y);
      }
      cletter=cletter->next;
   }
}

int clear_screen(int r, int g, int b) {
   Uint32 mypixel;

   mypixel = SDL_MapRGB(screen->format, r, g, b);
   SDL_FillRect(screen,NULL,mypixel);
   SDL_Flip(screen);
   return 1;
}

void load_images() {
   SDL_Surface *imagemap;
   char filename[200];
   int x;

   if (!(imagemap = SDL_LoadBMP("title.bmp")))  {
      fprintf(stderr,"Unable to open title.bmp\n");
      exit(1);
   }
   if (imagemap == NULL) {
      fprintf(stderr,"Unable to open title.bmp\n");
      exit(1);
   }
   if (imagemap->format->palette) {
      SDL_SetColors(screen, imagemap->format->palette->colors, 0,
         imagemap->format->palette->ncolors); 
      title_image = SDL_DisplayFormat(imagemap);
      SDL_FreeSurface(imagemap);
   } else {
      fprintf(stderr,"Unable to set Palette.\n");
      exit(1);
   }

   if (!(imagemap = SDL_LoadBMP("sndon.bmp"))) {
      fprintf(stderr,"Unable to open sndon.bmp\n");
      exit(1);
   }

   SDL_SetColorKey(imagemap, (SDL_SRCCOLORKEY|SDL_RLEACCEL),
      *(Uint8 *)imagemap->pixels);

   sound_on = SDL_DisplayFormat(imagemap);
   SDL_FreeSurface(imagemap);

   if (!(imagemap = SDL_LoadBMP("sndoff.bmp"))) {
      fprintf(stderr,"Unable to open sndoff.bmp\n");
      exit(1);
   }

   SDL_SetColorKey(imagemap, (SDL_SRCCOLORKEY|SDL_RLEACCEL),
      *(Uint8 *)imagemap->pixels);

   sound_off = SDL_DisplayFormat(imagemap);
   SDL_FreeSurface(imagemap);

   for(x=MAXSTATE-1;x>=0;x--) {
      sprintf(filename,"man%d.bmp",x);
      if (!(imagemap = SDL_LoadBMP(filename))) {
         fprintf(stderr,"Unable to open %s\n",filename);
         exit(1);
      }
      hangman[x] = SDL_DisplayFormat(imagemap);
      SDL_FreeSurface(imagemap);
   }
}

int init_display() {
   Uint16 width = 640;
   Uint16 height = 480;
   Uint8 bits = 8;
   int ptsize=30;
   char myfont[15];
   char mytitle[10];

   sprintf(myfont,"timrom.ttf");
   sprintf(mytitle,"HangMan");

   if (soundflag > 0) {
      if ( SDL_Init(SDL_INIT_EVERYTHING) < 0 ) {
         fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
         exit(2);
      }
   } else {
      if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
         fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
         exit(2);
      }
   }
   atexit(SDL_Quit);

   SDL_WM_SetCaption("Hangman by Kent Mein", "hangman");

   if ( TTF_Init() < 0 ) {
      fprintf(stderr, "Couldn't initialize TTF: %s\n",SDL_GetError());
      exit(2);
   }
   atexit(TTF_Quit);

   screen = SDL_SetVideoMode(width, height, bits, SDL_SWSURFACE|SDL_ANYFORMAT);
   if ( screen == NULL ) {
      fprintf(stderr, "Couldn't initialize SDL: %s\n",SDL_GetError());
      exit(2);
   }

   load_images();
   SDL_EnableUNICODE(1);

   font = TTF_OpenFont(myfont,ptsize);
   if (font == NULL) {
      fprintf(stderr,"Woo problem with font. %s\n",SDL_GetError());
      exit(2);
   }
   blank =  TTF_RenderText(font,"_",white,black);
   blackblank = TTF_RenderText(font,"_",black,black);
   return 1;
}

int init_sounds() {
   char mysound[255];
   int x,y, audio_channels=2;

   /* Open the audio device */
   if ( Mix_OpenAudio(22050, AUDIO_S16, audio_channels, 1024) < 0 ) {
      fprintf(stderr,
         "Warning: Couldn't set 11025 Hz 8-bit audio\n- Reason: %s\n",
         SDL_GetError());
      return -1;
   }
   for (x=0;x<NUMSOUNDS;x++) {
      sprintf(mysound,"good%d.wav",x);
      if (!(good_sounds[x] = Mix_LoadWAV(mysound))) {
         for (y=0;y<x;y++) Mix_FreeChunk(good_sounds[y]);
         return -1;
      }
      sprintf(mysound,"bad%d.wav",x);
      if (!(bad_sounds[x] = Mix_LoadWAV(mysound))) {
         for (y=0;y<x;y++) Mix_FreeChunk(bad_sounds[y]);
         return -1;
      }
   }
   return 1;
}

void do_end(struct word *myword,int type) {
   SDL_Surface *image;
   char msg[10];

   if (type == WINNER) {
      sprintf(msg,"You win!");
   } else {
      sprintf(msg,"You lose!");
   }

   clear_screen(1,1,1);
   image =  TTF_RenderText(font,msg,white,black);
   if (image == NULL) {
      fprintf(stderr,"Problem rendering font. %s\n",SDL_GetError());
      exit(2);
   }
   dest.x = 200;
   dest.y = 200 - image->h;
   dest.w = image->w;
   dest.h = image->h;
   if (SDL_BlitSurface(image, NULL, screen, &dest) < 0) {
      fprintf(stderr,"Couldn't blit text to display: %s\n",SDL_GetError());
      exit(2);
   }
   SDL_FreeSurface(image);
   SDL_UpdateRects(screen,1,&dest);
   draw_display(0,myword);
}

int do_title_page() {
   int done = 0;
   SDL_Event event;

   draw_image(title_image,0,0);
   if (soundflag == 1) {
      draw_image(sound_on,460,245);
   } else {
      draw_image(sound_off,475,257);
   }

   while (!done) {
   while ((SDL_PollEvent(&event) == 0)) { }
      switch(event.type) {
         case SDL_QUIT: 
            return 0;
            break;
         case SDL_KEYDOWN:
            if (event.key.keysym.sym == SDLK_ESCAPE) {
               return 0;
            } else if (event.key.keysym.sym == SDLK_s) {
               // Redraw title on dest
               dest.x = 475;
               dest.y = 257;
               dest.w = sound_off->w;
               dest.h = sound_off->h;
               SDL_BlitSurface(title_image, &dest, screen, &dest);
               SDL_UpdateRects(screen,1,&dest);

               if (soundflag == 1) {
                  draw_image(sound_off,475,257);
                  soundflag = 0;
               } else if (soundflag == 0) {
                  draw_image(sound_on,460,245);
                  soundflag = 1;
               } else {
                  draw_image(sound_off,475,257);
               }
            } else {
               done = 1;
            }
            break;
      }
   }
   return 1;
}

int main_loop(char *dictionary) {
   int state,winner,picked,which, sound_toggle;
   Uint16 newkey;
   SDL_Event event;
   struct word *myword, *current;

   if (!do_title_page()) return 1;
   if (soundflag == -2) return 1;
   state = winner = picked = 0;
   clear_screen(1,1,1);
   myword = get_word(dictionary, myword);
   if (myword == NULL)  {
      fprintf(stderr,"Error in reading dictionary.\n");
      return 1;
   }
   draw_display(state,myword);
   while ((state < MAXSTATE) && !winner) {
      while ((SDL_PollEvent(&event) == 0)) {
      }
      switch (event.type) {
         case SDL_QUIT:
            myfree(myword);
	    return 1;
            break;
         case SDL_KEYDOWN:
            if (event.key.keysym.sym == SDLK_ESCAPE) {
               myfree(myword);
               return 0;
            } else {
               newkey = event.key.keysym.unicode;
               current=myword;
               picked = 0;
               winner = 1;
               while (current!=NULL) {
                  if ((tolower(newkey) == current->letter) || 
	              (toupper(newkey) == current->letter) ) {
                     current->flag = 1;
                     picked = 1;
                  }
                  if (current->flag == 0) winner = 0;
   	          current= current->next;
               }
               if (soundflag > 0) {
                  which=(int)(rand() % NUMSOUNDS);
                  if (!picked) {
                     state++;
                     Mix_PlayChannel(0,bad_sounds[which],0);
                  } else {
 		     Mix_PlayChannel(0,good_sounds[which],0);
                  }
 	          while ( Mix_Playing(0) ) {
                     SDL_Delay(100);
                  }
               }
               draw_display(state,myword);
               if (soundflag) while (Mix_Playing(0)) { SDL_Delay(20); }
 	    }
            break;
      }
   }
   current = myword;
   while (current != NULL) {
      current->flag =1;
      current=current->next;
   }
   if (winner) do_end(myword,WINNER);
   else do_end(myword,LOOSER);
   while ( SDL_PollEvent(&event) == 0 && 
      (event.type !=SDL_QUIT) || (event.type !=SDL_KEYDOWN)) { }
   if (event.type == SDL_QUIT) {
      myfree(myword);
      return 1;
   } else {
      if (event.key.keysym.sym == SDLK_ESCAPE) {
         myfree(myword);
         return 1;
      }
   }
   myfree(myword);
   return 0;
}

int main(int argc, char **argv) {
   char dictionary[255] = "hangman.dat";
   int x;

   srand(time(NULL));
   soundflag=1;

   for (x=1;x<argc;x++) {
      if (strcmp("-nosound",argv[x]) == 0) {
         soundflag=0;
      } else {
         sprintf(dictionary, "%s", argv[x]);
      }
   }
   if (!init_display()) return 0;
   if (soundflag) soundflag = init_sounds();
   while (!main_loop(dictionary)) {}
   for(x=0;x<MAXSTATE;x++) SDL_FreeSurface(hangman[x]);
   SDL_FreeSurface(title_image);
   SDL_FreeSurface(sound_on);
   SDL_FreeSurface(sound_off);
   SDL_FreeSurface(blank);
   SDL_FreeSurface(blackblank);
   TTF_CloseFont(font);
   for (x=0;x<NUMSOUNDS;x++) {
      if (good_sounds[x] !=NULL) Mix_FreeChunk(good_sounds[x]);
      if (bad_sounds[x] !=NULL) Mix_FreeChunk(bad_sounds[x]);
   }
   fprintf(stderr,"Thank You for playing.\n");
   return(1);
}
